/*
 * SPDX-FileCopyrightText: 2023 jordi fita mas <jfita@peritasoft.com>
 * SPDX-License-Identifier: AGPL-3.0-only
 */

package types

import (
	"context"
	"fmt"
	"net/http"

	"github.com/jackc/pgx/v4"

	"dev.tandem.ws/tandem/camper/pkg/auth"
	"dev.tandem.ws/tandem/camper/pkg/database"
	"dev.tandem.ws/tandem/camper/pkg/form"
	httplib "dev.tandem.ws/tandem/camper/pkg/http"
	"dev.tandem.ws/tandem/camper/pkg/locale"
	"dev.tandem.ws/tandem/camper/pkg/template"
	"dev.tandem.ws/tandem/camper/pkg/uuid"
)

type AdminHandler struct {
	locales locale.Locales
}

func NewAdminHandler(locales locale.Locales) *AdminHandler {
	return &AdminHandler{locales}
}

func (h *AdminHandler) Handler(user *auth.User, company *auth.Company, conn *database.Conn) http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		var head string
		head, r.URL.Path = httplib.ShiftPath(r.URL.Path)

		switch head {
		case "new":
			switch r.Method {
			case http.MethodGet:
				f, err := newTypeForm(r.Context(), company, conn)
				if err != nil {
					panic(err)
				}
				f.MustRender(w, r, user, company)
			default:
				httplib.MethodNotAllowed(w, r, http.MethodGet)
			}
		case "":
			switch r.Method {
			case http.MethodGet:
				serveTypeIndex(w, r, user, company, conn)
			case http.MethodPost:
				addType(w, r, user, company, conn)
			default:
				httplib.MethodNotAllowed(w, r, http.MethodGet, http.MethodPost)
			}
		case "order":
			switch r.Method {
			case http.MethodPost:
				orderTypes(w, r, user, company, conn)
			default:
				httplib.MethodNotAllowed(w, r, http.MethodPost)
			}
		default:
			if !uuid.Valid(head) {
				http.NotFound(w, r)
				return
			}
			f, err := newTypeForm(r.Context(), company, conn)
			if err != nil {
				panic(err)
			}
			if err := f.FillFromDatabase(r.Context(), conn, head); err != nil {
				if database.ErrorIsNotFound(err) {
					http.NotFound(w, r)
					return
				}
				panic(err)
			}
			h.typeHandler(user, company, conn, f).ServeHTTP(w, r)
		}
	})
}

func (h *AdminHandler) typeHandler(user *auth.User, company *auth.Company, conn *database.Conn, f *typeForm) http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		var head string
		head, r.URL.Path = httplib.ShiftPath(r.URL.Path)

		switch head {
		case "":
			switch r.Method {
			case http.MethodGet:
				f.MustRender(w, r, user, company)
			case http.MethodPut:
				editType(w, r, user, company, conn, f)
			default:
				httplib.MethodNotAllowed(w, r, http.MethodGet, http.MethodPut)
			}
		case "features":
			h.featuresHandler(user, company, conn, f.Slug).ServeHTTP(w, r)
		case "options":
			h.optionsHandler(user, company, conn, f.Slug).ServeHTTP(w, r)
		case "slides":
			h.carouselHandler(user, company, conn, f.Slug).ServeHTTP(w, r)
		default:
			loc, ok := h.locales.Get(head)
			if !ok {
				http.NotFound(w, r)
				return
			}
			l10n := newTypeL10nForm(f, loc)
			if err := l10n.FillFromDatabase(r.Context(), conn); err != nil {
				panic(err)
			}
			switch r.Method {
			case http.MethodGet:
				l10n.MustRender(w, r, user, company)
			case http.MethodPut:
				editTypeL10n(w, r, user, company, conn, l10n)
			default:
				httplib.MethodNotAllowed(w, r, http.MethodGet, http.MethodPut)
			}
		}
	})
}

func serveTypeIndex(w http.ResponseWriter, r *http.Request, user *auth.User, company *auth.Company, conn *database.Conn) {
	types, err := collectTypeEntries(r.Context(), company, conn)
	if err != nil {
		panic(err)
	}
	page := &typeIndex{
		Types: types,
	}
	page.MustRender(w, r, user, company)
}

type typeEntry struct {
	Slug         string
	Name         string
	Active       bool
	Translations []*translation
}

type translation struct {
	Language string
	Endonym  string
	Missing  bool
}

func collectTypeEntries(ctx context.Context, company *auth.Company, conn *database.Conn) ([]*typeEntry, error) {
	rows, err := conn.Query(ctx, `
		select campsite_type.slug
		     , campsite_type.name
		     , campsite_type.active
		     , array_agg((lang_tag, endonym, not exists (select 1 from campsite_type_i18n as i18n where i18n.campsite_type_id = campsite_type.campsite_type_id and i18n.lang_tag = language.lang_tag)) order by endonym)
		from campsite_type
		    join company using (company_id)
		   , language
		where lang_tag <> default_lang_tag
		  and language.selectable
		  and campsite_type.company_id = $1
		group by campsite_type.slug
		       , campsite_type.name
		       , campsite_type.position
		       , campsite_type.active
		order by position, name
	`, pgx.QueryResultFormats{pgx.BinaryFormatCode}, company.ID)
	if err != nil {
		return nil, err
	}
	defer rows.Close()

	var types []*typeEntry
	for rows.Next() {
		entry := &typeEntry{}
		var translations database.RecordArray
		if err = rows.Scan(&entry.Slug, &entry.Name, &entry.Active, &translations); err != nil {
			return nil, err
		}
		for _, el := range translations.Elements {
			entry.Translations = append(entry.Translations, &translation{
				el.Fields[0].Get().(string),
				el.Fields[1].Get().(string),
				el.Fields[2].Get().(bool),
			})
		}
		types = append(types, entry)
	}

	return types, nil
}

type typeIndex struct {
	Types []*typeEntry
}

func (page *typeIndex) MustRender(w http.ResponseWriter, r *http.Request, user *auth.User, company *auth.Company) {
	template.MustRenderAdmin(w, r, user, company, "campsite/type/index.gohtml", page)
}

func addType(w http.ResponseWriter, r *http.Request, user *auth.User, company *auth.Company, conn *database.Conn) {
	f, err := newTypeForm(r.Context(), company, conn)
	if err != nil {
		panic(err)
	}
	processTypeForm(w, r, user, company, conn, f, func(ctx context.Context, tx *database.Tx) error {
		slug, err := tx.AddCampsiteType(ctx, company.ID, f.Media.Int(), f.Name.Val, f.Spiel.Val, f.Info.Val, f.Facilities.Val, f.Description.Val, f.MaxCampers.Int(), f.DogsAllowed.Checked)
		if err != nil {
			return err
		}
		return setTypePrices(ctx, tx, slug, f.Prices)
	})
}

func setTypePrices(ctx context.Context, tx *database.Tx, slug string, prices map[int]*typePriceForm) error {
	for seasonID, p := range prices {
		if err := tx.SetCampsiteTypeCost(ctx, slug, seasonID, p.MinNights.Int(), p.PricePerNight.Val); err != nil {
			return err
		}
	}
	return nil
}

func editType(w http.ResponseWriter, r *http.Request, user *auth.User, company *auth.Company, conn *database.Conn, f *typeForm) {
	processTypeForm(w, r, user, company, conn, f, func(ctx context.Context, tx *database.Tx) error {
		if _, err := tx.EditCampsiteType(ctx, f.Slug, f.Media.Int(), f.Name.Val, f.Spiel.Val, f.Info.Val, f.Facilities.Val, f.Description.Val, f.MaxCampers.Int(), f.DogsAllowed.Checked, f.Active.Checked); err != nil {
			return err
		}
		return setTypePrices(ctx, tx, f.Slug, f.Prices)
	})
}

func orderTypes(w http.ResponseWriter, r *http.Request, user *auth.User, company *auth.Company, conn *database.Conn) {
	if err := r.ParseForm(); err != nil {
		http.Error(w, err.Error(), http.StatusBadRequest)
		return
	}
	if err := user.VerifyCSRFToken(r); err != nil {
		http.Error(w, err.Error(), http.StatusForbidden)
		return
	}
	slugs := r.PostForm["slug"]
	if len(slugs) > 0 {
		for _, slug := range slugs {
			if !uuid.Valid(slug) {
				w.WriteHeader(http.StatusUnprocessableEntity)
				return
			}
		}
		if err := conn.OrderCampsiteTypes(r.Context(), slugs); err != nil {
			panic(err)
		}
	}
	serveTypeIndex(w, r, user, company, conn)
}

func processTypeForm(w http.ResponseWriter, r *http.Request, user *auth.User, company *auth.Company, conn *database.Conn, f *typeForm, act func(ctx context.Context, tx *database.Tx) error) {
	if err := f.Parse(r); err != nil {
		http.Error(w, err.Error(), http.StatusBadRequest)
		return
	}
	if err := user.VerifyCSRFToken(r); err != nil {
		http.Error(w, err.Error(), http.StatusForbidden)
		return
	}
	if ok, err := f.Valid(r.Context(), conn, user.Locale); err != nil {
		panic(err)
	} else if !ok {
		if !httplib.IsHTMxRequest(r) {
			w.WriteHeader(http.StatusUnprocessableEntity)
		}
		f.MustRender(w, r, user, company)
		return
	}

	tx := conn.MustBegin(r.Context())
	if err := act(r.Context(), tx); err == nil {
		if err := tx.Commit(r.Context()); err != nil {
			panic(err)
		}
	} else {
		if err := tx.Rollback(r.Context()); err != nil {
			panic(err)
		}
		panic(err)
	}
	httplib.Redirect(w, r, "/admin/campsites/types", http.StatusSeeOther)
}

type typeForm struct {
	Slug        string
	Active      *form.Checkbox
	Media       *form.Media
	Name        *form.Input
	MaxCampers  *form.Input
	DogsAllowed *form.Checkbox
	Spiel       *form.Input
	Info        *form.Input
	Facilities  *form.Input
	Description *form.Input
	Prices      map[int]*typePriceForm
}

type typePriceForm struct {
	SeasonName    string
	PricePerNight *form.Input
	MinNights     *form.Input
}

func newTypeForm(ctx context.Context, company *auth.Company, conn *database.Conn) (*typeForm, error) {
	f := &typeForm{
		Active: &form.Checkbox{
			Name:    "active",
			Checked: true,
		},
		Media: &form.Media{
			Input: &form.Input{
				Name: "media",
			},
			Label:  locale.PgettextNoop("Cover image", "input"),
			Prompt: locale.PgettextNoop("Set campsite type cover", "action"),
		},
		Name: &form.Input{
			Name: "name",
		},
		MaxCampers: &form.Input{
			Name: "max_campers",
		},
		DogsAllowed: &form.Checkbox{
			Name: "dogs_allowed",
		},
		Spiel: &form.Input{
			Name: "spiel",
		},
		Info: &form.Input{
			Name: "info",
		},
		Facilities: &form.Input{
			Name: "facilities",
		},
		Description: &form.Input{
			Name: "description",
		},
	}

	rows, err := conn.Query(ctx, "select season_id, name from season where active and company_id = $1", company.ID)
	if err != nil {
		return nil, err
	}
	defer rows.Close()

	f.Prices = make(map[int]*typePriceForm)
	for rows.Next() {
		var id int
		var name string
		if err = rows.Scan(&id, &name); err != nil {
			return nil, err
		}
		f.Prices[id] = &typePriceForm{
			SeasonName: name,
			PricePerNight: &form.Input{
				Name: fmt.Sprintf("season.%d.price_per_night", id),
				Val:  "0",
			},
			MinNights: &form.Input{
				Name: fmt.Sprintf("season.%d.min_nights", id),
				Val:  "1",
			},
		}
	}

	return f, nil
}

func (f *typeForm) FillFromDatabase(ctx context.Context, conn *database.Conn, slug string) error {
	f.Slug = slug
	row := conn.QueryRow(ctx, `
		select name
		     , spiel::text
		     , info::text
		     , facilities::text
		     , description::text
		     , media_id::text
		     , max_campers::text
		     , dogs_allowed
		     , active
		from campsite_type
		where slug = $1
	`, slug)
	if err := row.Scan(&f.Name.Val, &f.Spiel.Val, &f.Info.Val, &f.Facilities.Val, &f.Description.Val, &f.Media.Val, &f.MaxCampers.Val, &f.DogsAllowed.Checked, &f.Active.Checked); err != nil {
		return err
	}

	rows, err := conn.Query(ctx, `
		select season_id
		     , min_nights::text
		     , to_price(cost_per_night)
		from campsite_type
		join campsite_type_cost using (campsite_type_id)
		where slug = $1
	`, slug)
	if err != nil {
		return err
	}
	defer rows.Close()

	for rows.Next() {
		var seasonID int
		var minNights string
		var pricePerNight string
		if err := rows.Scan(&seasonID, &minNights, &pricePerNight); err != nil {
			return err
		}
		if p, ok := f.Prices[seasonID]; ok {
			p.MinNights.Val = minNights
			p.PricePerNight.Val = pricePerNight
		}
	}
	return rows.Err()
}

func (f *typeForm) Parse(r *http.Request) error {
	if err := r.ParseForm(); err != nil {
		return err
	}
	f.Active.FillValue(r)
	f.Name.FillValue(r)
	f.MaxCampers.FillValue(r)
	f.DogsAllowed.FillValue(r)
	f.Spiel.FillValue(r)
	f.Info.FillValue(r)
	f.Facilities.FillValue(r)
	f.Description.FillValue(r)
	f.Media.FillValue(r)
	for _, p := range f.Prices {
		p.PricePerNight.FillValue(r)
		p.MinNights.FillValue(r)
	}
	return nil
}

func (f *typeForm) Valid(ctx context.Context, conn *database.Conn, l *locale.Locale) (bool, error) {
	v := form.NewValidator(l)
	if v.CheckRequired(f.Name, l.GettextNoop("Name can not be empty.")) {
		v.CheckMinLength(f.Name, 1, l.GettextNoop("Name must have at least one letter."))
	}
	if v.CheckRequired(f.Media.Input, l.GettextNoop("Cover image can not be empty.")) {
		if _, err := v.CheckImageMedia(ctx, conn, f.Media.Input, l.GettextNoop("Cover image must be an image media type.")); err != nil {
			return false, err
		}
	}
	if v.CheckRequired(f.MaxCampers, l.GettextNoop("Maximum number of campers can not be empty.")) {
		if v.CheckValidInteger(f.MaxCampers, l.GettextNoop("Maximum number of campers must be an integer number.")) {
			v.CheckMinInteger(f.MaxCampers, 1, l.GettextNoop("Maximum number of campers must be one or greater."))
		}
	}
	for _, p := range f.Prices {
		if v.CheckRequired(p.PricePerNight, l.GettextNoop("Price per night can not be empty.")) {
			if v.CheckValidDecimal(p.PricePerNight, l.GettextNoop("Price per night must be a decimal number.")) {
				v.CheckMinDecimal(p.PricePerNight, 0.0, l.GettextNoop("Price per night must be zero or greater."))
			}
		}
		if v.CheckRequired(p.MinNights, l.GettextNoop("Minimum number of nights can not be empty.")) {
			if v.CheckValidInteger(p.MinNights, l.GettextNoop("Minimum number of nights must be an integer.")) {
				v.CheckMinInteger(p.MinNights, 0, l.GettextNoop("Minimum number of nights must be one or greater."))
			}
		}
	}
	return v.AllOK, nil
}

func (f *typeForm) MustRender(w http.ResponseWriter, r *http.Request, user *auth.User, company *auth.Company) {
	template.MustRenderAdmin(w, r, user, company, "campsite/type/form.gohtml", f)
}
