/*
 * SPDX-FileCopyrightText: 2023 jordi fita mas <jfita@peritasoft.com>
 * SPDX-License-Identifier: AGPL-3.0-only
 */

package types

import (
	"context"
	"net/http"
	"strconv"

	"github.com/jackc/pgx/v4"

	"dev.tandem.ws/tandem/camper/pkg/auth"
	"dev.tandem.ws/tandem/camper/pkg/carousel"
	"dev.tandem.ws/tandem/camper/pkg/database"
	"dev.tandem.ws/tandem/camper/pkg/form"
	httplib "dev.tandem.ws/tandem/camper/pkg/http"
	"dev.tandem.ws/tandem/camper/pkg/locale"
	"dev.tandem.ws/tandem/camper/pkg/template"
)

func (h *AdminHandler) carouselHandler(user *auth.User, company *auth.Company, conn *database.Conn, typeSlug string) http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		var head string
		head, r.URL.Path = httplib.ShiftPath(r.URL.Path)

		switch head {
		case "":
			switch r.Method {
			case http.MethodGet:
				serveCarouselIndex(w, r, user, company, conn, typeSlug)
			case http.MethodPost:
				addSlide(w, r, user, company, conn, typeSlug)
			default:
				httplib.MethodNotAllowed(w, r, http.MethodGet, http.MethodGet)
			}
		case "new":
			switch r.Method {
			case http.MethodGet:
				f := newSlideForm(typeSlug)
				f.MustRender(w, r, user, company)
			default:
				httplib.MethodNotAllowed(w, r, http.MethodGet)
			}
		case "order":
			switch r.Method {
			case http.MethodPost:
				orderCarousel(w, r, user, company, conn, typeSlug)
			default:
				httplib.MethodNotAllowed(w, r, http.MethodGet)
			}
		default:
			mediaID, err := strconv.Atoi(head)
			if err != nil {
				http.NotFound(w, r)
				return
			}
			f := newSlideForm(typeSlug)
			if err := f.FillFromDatabase(r.Context(), conn, mediaID); err != nil {
				if database.ErrorIsNotFound(err) {
					http.NotFound(w, r)
					return
				}
				panic(err)
			}

			var langTag string
			langTag, r.URL.Path = httplib.ShiftPath(r.URL.Path)

			switch langTag {
			case "":
				switch r.Method {
				case http.MethodGet:
					f.MustRender(w, r, user, company)
				case http.MethodPut:
					editSlide(w, r, user, company, conn, f)
				case http.MethodDelete:
					h.deleteSlide(w, r, user, conn, typeSlug, mediaID)
				default:
					httplib.MethodNotAllowed(w, r, http.MethodGet, http.MethodPut, http.MethodDelete)
				}
			default:
				loc, ok := h.locales.Get(langTag)
				if !ok {
					http.NotFound(w, r)
					return
				}
				l10n := newSlideL10nForm(f, loc)
				if err := l10n.FillFromDatabase(r.Context(), conn); err != nil {
					panic(err)
				}
				switch r.Method {
				case http.MethodGet:
					l10n.MustRender(w, r, user, company)
				case http.MethodPut:
					editSlideL10n(w, r, user, company, conn, l10n)
				default:
					httplib.MethodNotAllowed(w, r, http.MethodGet, http.MethodPut)
				}
			}
		}
	})
}

func serveCarouselIndex(w http.ResponseWriter, r *http.Request, user *auth.User, company *auth.Company, conn *database.Conn, typeSlug string) {
	slides, err := collectSlideEntries(r.Context(), conn, typeSlug)
	if err != nil {
		panic(err)
	}
	page := &carouselIndex{
		TypeSlug: typeSlug,
		Slides:   slides,
	}
	page.MustRender(w, r, user, company)
}

type carouselIndex struct {
	TypeSlug string
	Slides   []*carousel.SlideEntry
}

func (page *carouselIndex) MustRender(w http.ResponseWriter, r *http.Request, user *auth.User, company *auth.Company) {
	template.MustRenderAdmin(w, r, user, company, "campsite/carousel/index.gohtml", page)
}

func mustCollectSlides(ctx context.Context, conn *database.Conn, loc *locale.Locale, typeSlug string) []*carousel.Slide {
	rows, err := conn.Query(ctx, `
		select coalesce(i18n.caption, slide.caption) as l10_caption
		     , media.path
		from campsite_type_carousel as slide
		    join campsite_type using (campsite_type_id)
		    join media on media.media_id = slide.media_id
		    left join campsite_type_carousel_i18n as i18n
		        on i18n.campsite_type_id = slide.campsite_type_id
		        and i18n.media_id = slide.media_id
				and lang_tag = $1
		where campsite_type.slug = $2
		order by slide.position, l10_caption
	`, loc.Language, typeSlug)
	if err != nil {
		panic(err)
	}
	defer rows.Close()

	var slides []*carousel.Slide
	for rows.Next() {
		slide := &carousel.Slide{}
		err = rows.Scan(&slide.Caption, &slide.Media)
		if err != nil {
			panic(err)
		}
		slides = append(slides, slide)
	}
	if rows.Err() != nil {
		panic(rows.Err())
	}

	return slides
}

func collectSlideEntries(ctx context.Context, conn *database.Conn, typeSlug string) ([]*carousel.SlideEntry, error) {
	rows, err := conn.Query(ctx, `
		select carousel.media_id
		     , media.path
		     , caption
		     , array_agg((lang_tag, endonym, not exists (select 1 from campsite_type_carousel_i18n as i18n where i18n.media_id = carousel.media_id and i18n.lang_tag = language.lang_tag)) order by endonym)
		from campsite_type_carousel as carousel
		    join campsite_type using (campsite_type_id)
		    join media on media.media_id = carousel.media_id
		    join company on company.company_id = campsite_type.company_id
		   , language
		where lang_tag <> default_lang_tag
		  and language.selectable
		  and campsite_type.slug = $1
		group by carousel.media_id
		       , carousel.position
		       , media.path
		       , caption
		order by carousel.position, caption
	`, pgx.QueryResultFormats{pgx.BinaryFormatCode}, typeSlug)
	if err != nil {
		return nil, err
	}
	defer rows.Close()

	var slides []*carousel.SlideEntry
	for rows.Next() {
		slide := &carousel.SlideEntry{}
		var translations database.RecordArray
		if err = rows.Scan(&slide.ID, &slide.Media, &slide.Caption, &translations); err != nil {
			return nil, err
		}
		for _, el := range translations.Elements {
			slide.Translations = append(slide.Translations, &carousel.Translation{
				Language: el.Fields[0].Get().(string),
				Endonym:  el.Fields[1].Get().(string),
				Missing:  el.Fields[2].Get().(bool),
			})
		}
		slides = append(slides, slide)
	}

	return slides, nil
}

func addSlide(w http.ResponseWriter, r *http.Request, user *auth.User, company *auth.Company, conn *database.Conn, typeSlug string) {
	f := newSlideForm(typeSlug)
	editSlide(w, r, user, company, conn, f)
}

func editSlide(w http.ResponseWriter, r *http.Request, user *auth.User, company *auth.Company, conn *database.Conn, f *slideForm) {
	f.process(w, r, user, company, conn, func(ctx context.Context) {
		conn.MustExec(ctx, "select add_campsite_type_carousel_slide($1, $2, $3)", f.TypeSlug, f.Media, f.Caption)
	})
}

func (h *AdminHandler) deleteSlide(w http.ResponseWriter, r *http.Request, user *auth.User, conn *database.Conn, typeSlug string, mediaId int) {
	if err := user.VerifyCSRFToken(r); err != nil {
		http.Error(w, err.Error(), http.StatusForbidden)
		return
	}
	conn.MustExec(r.Context(), "select remove_campsite_type_carousel_slide($1, $2)", typeSlug, mediaId)
	httplib.Redirect(w, r, "/admin/campsites/types/"+typeSlug+"/slides", http.StatusSeeOther)
}

type slideForm struct {
	TypeSlug string
	MediaID  int
	Media    *form.Media
	Caption  *form.Input
}

func newSlideForm(typeSlug string) *slideForm {
	return &slideForm{
		TypeSlug: typeSlug,
		Media: &form.Media{
			Input: &form.Input{
				Name: "media",
			},
			Label:  locale.PgettextNoop("Slide image", "input"),
			Prompt: locale.PgettextNoop("Set slide image", "action"),
		},
		Caption: &form.Input{
			Name: "caption",
		},
	}
}

func (f *slideForm) FillFromDatabase(ctx context.Context, conn *database.Conn, mediaID int) error {
	f.MediaID = mediaID
	row := conn.QueryRow(ctx, `
		select caption
		     , carousel.media_id::text
		from campsite_type_carousel as carousel
		join campsite_type using (campsite_type_id)
		where campsite_type.slug = $1
		  and carousel.media_id = $2
	`, f.TypeSlug, mediaID)
	return row.Scan(&f.Caption.Val, &f.Media.Val)
}

func (f *slideForm) process(w http.ResponseWriter, r *http.Request, user *auth.User, company *auth.Company, conn *database.Conn, act func(ctx context.Context)) {
	if err := f.Parse(r); err != nil {
		http.Error(w, err.Error(), http.StatusBadRequest)
		return
	}
	if err := user.VerifyCSRFToken(r); err != nil {
		http.Error(w, err.Error(), http.StatusForbidden)
		return
	}
	if ok, err := f.Valid(r.Context(), conn, user.Locale); err != nil {
		panic(err)
	} else if !ok {
		if !httplib.IsHTMxRequest(r) {
			w.WriteHeader(http.StatusUnprocessableEntity)
		}
		f.MustRender(w, r, user, company)
		return
	}
	act(r.Context())
	httplib.Redirect(w, r, "/admin/campsites/types/"+f.TypeSlug+"/slides", http.StatusSeeOther)
}

func (f *slideForm) Parse(r *http.Request) error {
	if err := r.ParseForm(); err != nil {
		return err
	}
	f.Caption.FillValue(r)
	f.Media.FillValue(r)
	return nil
}

func (f *slideForm) Valid(ctx context.Context, conn *database.Conn, l *locale.Locale) (bool, error) {
	v := form.NewValidator(l)
	if v.CheckRequired(f.Media.Input, l.GettextNoop("Slide image can not be empty.")) {
		if _, err := v.CheckImageMedia(ctx, conn, f.Media.Input, l.GettextNoop("Slide image must be an image media type.")); err != nil {
			return false, err
		}
	}
	return v.AllOK, nil
}

func (f *slideForm) MustRender(w http.ResponseWriter, r *http.Request, user *auth.User, company *auth.Company) {
	template.MustRenderAdmin(w, r, user, company, "campsite/carousel/form.gohtml", f)
}

func orderCarousel(w http.ResponseWriter, r *http.Request, user *auth.User, company *auth.Company, conn *database.Conn, typeSlug string) {
	if err := r.ParseForm(); err != nil {
		http.Error(w, err.Error(), http.StatusBadRequest)
		return
	}
	if err := user.VerifyCSRFToken(r); err != nil {
		http.Error(w, err.Error(), http.StatusForbidden)
		return
	}
	input := r.PostForm["media_id"]
	if len(input) > 0 {
		var ids []int
		for _, s := range input {
			if id, err := strconv.Atoi(s); err == nil {
				ids = append(ids, id)
			} else {
				http.Error(w, err.Error(), http.StatusUnprocessableEntity)
				return
			}
		}
		if err := conn.OrderCampsiteTypeCarousel(r.Context(), typeSlug, ids); err != nil {
			panic(err)
		}
	}
	serveCarouselIndex(w, r, user, company, conn, typeSlug)
}
