/*
 * SPDX-FileCopyrightText: 2023 jordi fita mas <jfita@peritasoft.com>
 * SPDX-License-Identifier: AGPL-3.0-only
 */

package types

import (
	"context"
	"net/http"

	"dev.tandem.ws/tandem/camper/pkg/auth"
	"dev.tandem.ws/tandem/camper/pkg/database"
	"dev.tandem.ws/tandem/camper/pkg/form"
	httplib "dev.tandem.ws/tandem/camper/pkg/http"
	"dev.tandem.ws/tandem/camper/pkg/locale"
	"dev.tandem.ws/tandem/camper/pkg/template"
)

type typeL10nForm struct {
	Locale      *locale.Locale
	Slug        string
	Name        *form.L10nInput
	Spiel       *form.L10nInput
	Info        *form.L10nInput
	Facilities  *form.L10nInput
	Description *form.L10nInput
}

func newTypeL10nForm(f *typeForm, loc *locale.Locale) *typeL10nForm {
	return &typeL10nForm{
		Locale:      loc,
		Slug:        f.Slug,
		Name:        f.Name.L10nInput(),
		Spiel:       f.Spiel.L10nInput(),
		Info:        f.Info.L10nInput(),
		Facilities:  f.Facilities.L10nInput(),
		Description: f.Description.L10nInput(),
	}
}

func (l10n *typeL10nForm) FillFromDatabase(ctx context.Context, conn *database.Conn) error {
	row := conn.QueryRow(ctx, `
		select coalesce(i18n.name, '') as l10n_name
		     , coalesce(i18n.spiel, '') as l10n_spiel
		     , coalesce(i18n.info, '') as l10n_info
		     , coalesce(i18n.facilities, '') as l10n_facilities
		     , coalesce(i18n.description, '') as l10n_description
		from campsite_type
		    left join campsite_type_i18n as i18n on campsite_type.campsite_type_id = i18n.campsite_type_id and i18n.lang_tag = $1
		where slug = $2
	`, l10n.Locale.Language, l10n.Slug)
	return row.Scan(&l10n.Name.Val, &l10n.Spiel.Val, &l10n.Info.Val, &l10n.Facilities.Val, &l10n.Description.Val)
}

func (l10n *typeL10nForm) MustRender(w http.ResponseWriter, r *http.Request, user *auth.User, company *auth.Company) {
	template.MustRenderAdmin(w, r, user, company, "campsite/type/l10n.gohtml", l10n)
}

func editTypeL10n(w http.ResponseWriter, r *http.Request, user *auth.User, company *auth.Company, conn *database.Conn, l10n *typeL10nForm) {
	if ok, err := form.Handle(l10n, w, r, user); err != nil {
		return
	} else if !ok {
		l10n.MustRender(w, r, user, company)
		return
	}
	if err := conn.TranslateCampsiteType(r.Context(), l10n.Slug, l10n.Locale.Language, l10n.Name.Val, l10n.Spiel.Val, l10n.Info.Val, l10n.Facilities.Val, l10n.Description.Val); err != nil {
		panic(err)
	}
	httplib.Redirect(w, r, "/admin/campsites/types", http.StatusSeeOther)
}

func (l10n *typeL10nForm) Parse(r *http.Request) error {
	if err := r.ParseForm(); err != nil {
		return err
	}
	l10n.Name.FillValue(r)
	l10n.Spiel.FillValue(r)
	l10n.Info.FillValue(r)
	l10n.Facilities.FillValue(r)
	l10n.Description.FillValue(r)
	return nil
}

func (l10n *typeL10nForm) Valid(l *locale.Locale) bool {
	v := form.NewValidator(l)
	v.CheckRequired(&l10n.Name.Input, l.GettextNoop("Name can not be empty."))
	return v.AllOK
}

type optionL10nForm struct {
	Locale   *locale.Locale
	TypeSlug string
	ID       int
	Name     *form.L10nInput
}

func newOptionL10nForm(f *optionForm, loc *locale.Locale) *optionL10nForm {
	return &optionL10nForm{
		Locale:   loc,
		TypeSlug: f.TypeSlug,
		ID:       f.ID,
		Name:     f.Name.L10nInput(),
	}
}

func (l10n *optionL10nForm) FillFromDatabase(ctx context.Context, conn *database.Conn) error {
	row := conn.QueryRow(ctx, `
		select coalesce(i18n.name, '') as l10n_name
		from campsite_type_option
		    left join campsite_type_option_i18n as i18n on campsite_type_option.campsite_type_option_id = i18n.campsite_type_option_id and i18n.lang_tag = $1
		where campsite_type_option.campsite_type_option_id = $2
	`, l10n.Locale.Language, l10n.ID)
	return row.Scan(&l10n.Name.Val)
}

func (l10n *optionL10nForm) MustRender(w http.ResponseWriter, r *http.Request, user *auth.User, company *auth.Company) {
	template.MustRenderAdmin(w, r, user, company, "campsite/option/l10n.gohtml", l10n)
}

func editOptionL10n(w http.ResponseWriter, r *http.Request, user *auth.User, company *auth.Company, conn *database.Conn, l10n *optionL10nForm) {
	if ok, err := form.Handle(l10n, w, r, user); err != nil {
		return
	} else if !ok {
		l10n.MustRender(w, r, user, company)
		return
	}
	if err := conn.TranslateCampsiteTypeOption(r.Context(), l10n.ID, l10n.Locale.Language, l10n.Name.Val); err != nil {
		panic(err)
	}
	httplib.Redirect(w, r, "/admin/campsites/types/"+l10n.TypeSlug+"/options", http.StatusSeeOther)
}

func (l10n *optionL10nForm) Parse(r *http.Request) error {
	if err := r.ParseForm(); err != nil {
		return err
	}
	l10n.Name.FillValue(r)
	return nil
}

func (l10n *optionL10nForm) Valid(l *locale.Locale) bool {
	v := form.NewValidator(l)
	v.CheckRequired(&l10n.Name.Input, l.GettextNoop("Name can not be empty."))
	return v.AllOK
}

type slideL10nForm struct {
	Locale   *locale.Locale
	TypeSlug string
	MediaID  int
	Caption  *form.L10nInput
}

func newSlideL10nForm(f *slideForm, loc *locale.Locale) *slideL10nForm {
	return &slideL10nForm{
		Locale:   loc,
		TypeSlug: f.TypeSlug,
		MediaID:  f.MediaID,
		Caption:  f.Caption.L10nInput(),
	}
}

func (l10n *slideL10nForm) FillFromDatabase(ctx context.Context, conn *database.Conn) error {
	row := conn.QueryRow(ctx, `
		select coalesce(i18n.caption, '') as l10n_caption
		from campsite_type_carousel as carousel
		    join campsite_type using (campsite_type_id)
		    left join campsite_type_carousel_i18n as i18n
		        on carousel.campsite_type_id = i18n.campsite_type_id
		        and carousel.media_id = i18n.media_id
		        and i18n.lang_tag = $1
		where campsite_type.slug = $2 and carousel.media_id = $3
	`, l10n.Locale.Language, l10n.TypeSlug, l10n.MediaID)
	return row.Scan(&l10n.Caption.Val)
}

func (l10n *slideL10nForm) MustRender(w http.ResponseWriter, r *http.Request, user *auth.User, company *auth.Company) {
	template.MustRenderAdmin(w, r, user, company, "campsite/carousel/l10n.gohtml", l10n)
}

func editSlideL10n(w http.ResponseWriter, r *http.Request, user *auth.User, company *auth.Company, conn *database.Conn, l10n *slideL10nForm) {
	if err := l10n.Parse(r); err != nil {
		http.Error(w, err.Error(), http.StatusBadRequest)
		return
	}
	if err := user.VerifyCSRFToken(r); err != nil {
		http.Error(w, err.Error(), http.StatusForbidden)
		return
	}
	if !l10n.Valid(user.Locale) {
		if !httplib.IsHTMxRequest(r) {
			w.WriteHeader(http.StatusUnprocessableEntity)
		}
		l10n.MustRender(w, r, user, company)
		return
	}
	conn.MustExec(r.Context(), "select translate_campsite_type_carousel_slide($1, $2, $3, $4)", l10n.TypeSlug, l10n.MediaID, l10n.Locale.Language, l10n.Caption)
	httplib.Redirect(w, r, "/admin/campsites/types/"+l10n.TypeSlug+"/slides", http.StatusSeeOther)
}

func (l10n *slideL10nForm) Parse(r *http.Request) error {
	if err := r.ParseForm(); err != nil {
		return err
	}
	l10n.Caption.FillValue(r)
	return nil
}

func (l10n *slideL10nForm) Valid(l *locale.Locale) bool {
	v := form.NewValidator(l)
	return v.AllOK
}

type featureL10nForm struct {
	Locale   *locale.Locale
	TypeSlug string
	ID       int
	Name     *form.L10nInput
}

func newFeatureL10nForm(f *featureForm, loc *locale.Locale) *featureL10nForm {
	return &featureL10nForm{
		Locale:   loc,
		TypeSlug: f.TypeSlug,
		ID:       f.ID,
		Name:     f.Name.L10nInput(),
	}
}

func (l10n *featureL10nForm) FillFromDatabase(ctx context.Context, conn *database.Conn) error {
	row := conn.QueryRow(ctx, `
		select coalesce(i18n.name, '') as l10n_name
		from campsite_type_feature
		    left join campsite_type_feature_i18n as i18n on campsite_type_feature.campsite_type_feature_id = i18n.campsite_type_feature_id and i18n.lang_tag = $1
		where campsite_type_feature.campsite_type_feature_id = $2
	`, l10n.Locale.Language, l10n.ID)
	return row.Scan(&l10n.Name.Val)
}

func (l10n *featureL10nForm) MustRender(w http.ResponseWriter, r *http.Request, user *auth.User, company *auth.Company) {
	template.MustRenderAdmin(w, r, user, company, "campsite/feature/l10n.gohtml", l10n)
}

func editFeatureL10n(w http.ResponseWriter, r *http.Request, user *auth.User, company *auth.Company, conn *database.Conn, l10n *featureL10nForm) {
	if ok, err := form.Handle(l10n, w, r, user); err != nil {
		return
	} else if !ok {
		l10n.MustRender(w, r, user, company)
		return
	}
	if err := conn.TranslateCampsiteTypeFeature(r.Context(), l10n.ID, l10n.Locale.Language, l10n.Name.Val); err != nil {
		panic(err)
	}
	httplib.Redirect(w, r, "/admin/campsites/types/"+l10n.TypeSlug+"/features", http.StatusSeeOther)
}

func (l10n *featureL10nForm) Parse(r *http.Request) error {
	if err := r.ParseForm(); err != nil {
		return err
	}
	l10n.Name.FillValue(r)
	return nil
}

func (l10n *featureL10nForm) Valid(l *locale.Locale) bool {
	v := form.NewValidator(l)
	v.CheckRequired(&l10n.Name.Input, l.GettextNoop("Name can not be empty."))
	return v.AllOK
}
