/*
 * SPDX-FileCopyrightText: 2023 jordi fita mas <jfita@peritasoft.com>
 * SPDX-License-Identifier: AGPL-3.0-only
 */

package season

import (
	"context"
	"net/http"

	"dev.tandem.ws/tandem/camper/pkg/auth"
	"dev.tandem.ws/tandem/camper/pkg/database"
	"dev.tandem.ws/tandem/camper/pkg/form"
	httplib "dev.tandem.ws/tandem/camper/pkg/http"
	"dev.tandem.ws/tandem/camper/pkg/locale"
	"dev.tandem.ws/tandem/camper/pkg/template"
)

type seasonL10nForm struct {
	Locale *locale.Locale
	Slug   string
	Name   *form.L10nInput
}

func newSeasonL10nForm(f *seasonForm, loc *locale.Locale) *seasonL10nForm {
	return &seasonL10nForm{
		Locale: loc,
		Slug:   f.Slug,
		Name:   f.Name.L10nInput(),
	}
}

func (l10n *seasonL10nForm) FillFromDatabase(ctx context.Context, conn *database.Conn) error {
	row := conn.QueryRow(ctx, `
		select coalesce(i18n.name, '') as l10n_name
		from season
		    left join season_i18n as i18n on season.season_id = i18n.season_id and i18n.lang_tag = $1
		where slug = $2
	`, l10n.Locale.Language, l10n.Slug)
	return row.Scan(&l10n.Name.Val)
}

func (l10n *seasonL10nForm) MustRender(w http.ResponseWriter, r *http.Request, user *auth.User, company *auth.Company) {
	template.MustRenderAdmin(w, r, user, company, "season/l10n.gohtml", l10n)
}

func editSeasonL10n(w http.ResponseWriter, r *http.Request, user *auth.User, company *auth.Company, conn *database.Conn, l10n *seasonL10nForm) {
	if ok, err := form.Handle(l10n, w, r, user); err != nil {
		return
	} else if !ok {
		l10n.MustRender(w, r, user, company)
		return
	}
	conn.MustExec(r.Context(), "select translate_season($1, $2, $3)", l10n.Slug, l10n.Locale.Language, l10n.Name)
	httplib.Redirect(w, r, "/admin/seasons", http.StatusSeeOther)
}

func (l10n *seasonL10nForm) Parse(r *http.Request) error {
	if err := r.ParseForm(); err != nil {
		return err
	}
	l10n.Name.FillValue(r)
	return nil
}

func (l10n *seasonL10nForm) Valid(l *locale.Locale) bool {
	v := form.NewValidator(l)
	v.CheckRequired(&l10n.Name.Input, l.GettextNoop("Name can not be empty."))
	return v.AllOK
}
